module cordic_apb_wrapper_unfolded
#(
  parameter integer unsigned DEPTH = 10
 )
(
  // Group: APB interface, clock and reset
  // Inputs:
  // Outputs: 
  input logic                 CLK,              // System Clock
  input logic                 RSTn,             // System Reset
  input logic [31:0]          PADDR,            // APB Address bus
  input logic                 PENABLE,          // APB bus enable
  input logic                 PSEL,             // APB slave select
  input logic [31:0]          PWDATA,           // APB write databus
  input logic                 PWRITE,           // APB write transfer
  output logic [31:0]         PRDATA,           // APB read databus 
  output logic                PREADY,           // APB ready (always 1)
  output logic                PSLVERR,          // APB error (always 0)
  // Group: enable input (from cordic last stage)
  // Inputs:
  //    en_i - enable coming from cordic
  input logic                 en_i,
  // Group: enable, x-y-z data outputs
  // Outputs:
  //    en_o - enable output to cordic
  //    data - data to cordic inputs
  `ifdef CORDIC_16
  output shortint            X_o,  
  output shortint            Y_o,  
  input shortint             data_in_cos, 
  input shortint             data_in_sin, 
  `elsif CORDIC_8
  output byte                X_o,  
  output byte                Y_o, 
  input byte                 data_in_cos,  
  input byte                 data_in_sin, 
  `else
  output int                 X_o,  
  output int                 Y_o,
  input int                  data_in_cos,
  input int                  data_in_sin,
  `endif
  output logic [31:0]        Z_o,
  output logic               en_o,
  output logic               operation_o 
);

//apb read 
function automatic is_read();
   return PENABLE & PSEL & ~PWRITE;
endfunction  

//apb write 
function automatic is_write();
   return PENABLE & PSEL & PWRITE;
endfunction 

//cordic registers
//0x00 = control
//0x04 = status
//0x08 = x_data
//0x0C = y_data
//0x10 = z_data
//0x14 = data_cos
//0x18 = data_sin
//logic [31:0] cordic_regs [6:0];
logic [31:0] cordic_regs [6:2];

//no error
//always ready
assign PREADY = 1'b1;
assign PSLVERR = 1'b0;

//input to state machine
logic state_input_rot, state_input_vect, state_clear;
assign state_input_rot = PENABLE & PSEL & PWRITE & PWDATA[0] & (PADDR == 32'h00000000);
assign state_input_vect = PENABLE & PSEL & PWRITE & PWDATA[1] & (PADDR == 32'h00000000);
assign state_clear = PENABLE & PSEL & PWRITE & PWDATA[1] & (PADDR == 32'h00000004);

//state machine 
parameter IDLE  = 3'b000,
          START_ROT = 3'b001,
          START_VECT = 3'b010,
          RUN   = 3'b011,
          DONE  = 3'b100;
         
logic [2:0] state, next_state;
logic out_rot, out_vect, sample, done;

integer counter;

always_comb //(state, state_input, state_clear)
begin
  case(state)
    IDLE : 
    begin
      if(state_input_rot == 1'b1)
        next_state = START_ROT;
      else if(state_input_vect == 1'b1)
        next_state = START_VECT;
      else
        next_state = IDLE;
    end
    START_ROT : 
    begin
      next_state = RUN;
    end 
    START_VECT : 
    begin
      next_state = RUN;
    end 
    RUN : 
    begin 
      if(counter == DEPTH)
      begin
        next_state = DONE;
      end
      else
      begin 
        next_state = RUN;
      end
    end 
    DONE : 
    begin
      if(state_clear == 1'b1)
        next_state = IDLE;
      else
        next_state = DONE;
    end
    default :
      next_state = IDLE;
    endcase
end

always_ff @(posedge CLK or negedge RSTn)
begin
  if (RSTn == 1'b0)
  begin
    counter  <= 6'd0;
    out_rot  <= 1'b0;
    out_vect <= 1'b0;
    sample   <= 1'b0;
    done     <=1'b0;
  end
  //else if(state == IDLE)
  else
  case(state)
  IDLE :
  begin
    counter  <= 6'd0;
    out_rot  <= 1'b0;
    out_vect <= 1'b0;
    sample   <= 1'b0;
    done     <= 1'b0;
  end
  //else if(state == START_ROT)
  START_ROT :
  begin
    counter  <= 6'd0;
    out_rot  <= 1'b1;
    out_vect <= 1'b0;
    sample   <= 1'b0;
    done     <= 1'b0;
  end
  //else if(state == START_VECT)
  START_VECT :
  begin
    counter  <= 6'd0;
    out_rot  <= 1'b0;
    out_vect <= 1'b1;
    sample   <= 1'b0;
    done     <= 1'b0;
  end
  //else if(state == RUN)
  RUN :
  begin
    counter  <= counter + 1;
    out_rot  <= 1'b0;
    out_vect <= 1'b0;
    sample   <= 1'b0;
    done     <= 1'b0;
  end
  //else if(state == DONE)
  DONE :
  begin
    counter  <= 6'd0;
    out_rot  <= 1'b0;
    out_vect <= 1'b0;
    sample   <= 1'b1;
    done     <= 1'b1;
  end
  //else 
  //begin
  //  counter  <= 6'd0;
  //  out_rot  <= 1'b0;
  //  out_vect <= 1'b0;
  //  sample   <= 1'b0;
  //  done     <= 1'b0;
  //end
  default : 
  begin
    counter  <= 6'd0;
    out_rot  <= 1'b0;
    out_vect <= 1'b0;
    sample   <= 1'b0;
    done     <= 1'b0;
  end
  endcase
end

always_ff @(posedge CLK or negedge RSTn)
begin
  if (RSTn == 1'b0)
    state <= IDLE;
  else
    state <= next_state;
end

//write process
always_ff @(posedge CLK or negedge RSTn)
  if(!RSTn)
  begin
    //cordic_regs = '{default:32'h00000000};
    //cordic_regs[0] <= 32'h00000000;
    //cordic_regs[1] <= 32'h00000000;
    cordic_regs[2] <= 32'h00000000;
    cordic_regs[3] <= 32'h00000000;
    cordic_regs[4] <= 32'h00000000;
  end
  else if(is_write())
  begin
     //if(PADDR == 32'h00000000)
     //begin
     //  cordic_regs[0][31:1] = PWDATA[31:2];
       //cordic_regs[0][0] = out;
     //end 
     //else if(PADDR == 32'h00000004)
     //begin
       //cordic_regs[1][0] = done;
     //end
     //else 
     if(PADDR == 32'h00000008)
     begin
       //x_data
       cordic_regs[2] <= PWDATA;
     end
     else if(PADDR == 32'h0000000C)
     begin
       //y_data
       cordic_regs[3] <= PWDATA;
     end
     else if(PADDR == 32'h00000010)
     begin
       //z_data
       cordic_regs[4] <= PWDATA;
     end
   end
   //else if(is_read())
   //begin
   //  if(PADDR == 32'h00000014)
   //  begin
   //    cordic_regs[5] = 32'h00000000;
   //  end  
   //  else if(PADDR == 32'h00000018)
   //  begin
   //    cordic_regs[6] = 32'h00000000;
   //  end
   //end
   //else if(sample == 1'b1)
   //begin
   //  cordic_regs[5] = data_in_cos;
   //  cordic_regs[6] = data_in_sin;
   //end
   
always_ff @(posedge CLK or negedge RSTn)
  //if(~RSTn)
  if(!RSTn)
  begin
    cordic_regs[5] <= 32'h00000000;
    cordic_regs[6] <= 32'h00000000;
  end   
  else
  if(sample == 1'b1)
  begin
    cordic_regs[5] <= data_in_cos;
    cordic_regs[6] <= data_in_sin;
  end 
   
//cordic_regs read
always_comb 
   if(is_read())
   begin
     if(PADDR == 32'h00000000)
     begin
       //PRDATA = cordic_regs[0];
       PRDATA = {30'h00000000,out_vect,out_rot};
     end 
     else if(PADDR == 32'h00000004)
     begin
       //PRDATA = cordic_regs[1];
       PRDATA = {31'h00000000,done};
     end
     else if(PADDR == 32'h00000008)
     begin
       PRDATA = cordic_regs[2];
     end
     else if(PADDR == 32'h0000000C)
     begin
       PRDATA = cordic_regs[3];
     end
     else if(PADDR == 32'h00000010)
     begin
       PRDATA = cordic_regs[4];
     end
     else if(PADDR == 32'h00000014)
     begin
       PRDATA = cordic_regs[5];
     end
     else if(PADDR == 32'h00000018)
     begin
       PRDATA = cordic_regs[6];
     end
     else
     begin
       PRDATA = 32'h00000000;
     end
   end
   else
   begin
       PRDATA = 32'h00000000;
   end

//output assignment
assign X_o = cordic_regs[2];
assign Y_o = cordic_regs[3];
assign Z_o = cordic_regs[4];
assign en_o = out_rot || out_vect;
assign operation_o = out_vect;

endmodule
