module cordic_stage_folded
#(
  parameter integer unsigned WIDTH = 32,
  parameter integer unsigned DEPTH = 10
)
(
  // Group: clk and enable inputs
  // Inputs:
  //    clk - clock signal
  //    rst - reset signal
  input logic                clk,  
  //input logic                en,
  input logic                op,
  input logic                rst,
  //output logic               en_o,
  // Group: X, Y and Z inputs
  // Inputs:
  //    X - x position used for cosine
  //    Y - y position used for sine
  //    Z - used to determine the next rotation
  `ifdef CORDIC_16
  input shortint             X_i,  
  input shortint             Y_i,
  `elsif CORDIC_8
  input byte                 X_i,  
  input byte                 Y_i,
  `else
  input int                  X_i,  
  input int                  Y_i,
  `endif
  input logic [WIDTH-1:0]    Z_i,
  // Group: X, Y and Z outputs
  // Outputs:
  //    X - x position to post processing
  //    Y - y position to post processing
  //    Z - z to post processing
  `ifdef CORDIC_16
  output shortint            X_o,  
  output shortint            Y_o, 
  `elsif CORDIC_8
  output byte                X_o,  
  output byte                Y_o, 
  `else
  output int                 X_o,  
  output int                 Y_o,
  `endif
  output logic [WIDTH-1:0]   Z_o,
  // Group: apb_wrapper state machine inputs
  // Outputs:
  //    en_reg - enable to registers
  //    counter - counter
  //    step - X_i or X_next
  //    conv - convergence step or normal
  input logic                en_reg,
  input integer              counter,
  input logic                step,
  input logic                conv   
);

//////////////////////////////////////////////////////////////////
/////////////////////LOOK UP TABLE////////////////////////////////
//////////////////////////////////////////////////////////////////
logic [31:0] angle_rom [20:0];

always_ff @(posedge clk or negedge rst)
if(!rst)
  begin
    angle_rom[0] <= 32'h20000000;
    angle_rom[1] <= 32'h12E4051E;
    angle_rom[2] <= 32'h09FB385B;
    angle_rom[3] <= 32'h051111D4;
    angle_rom[4] <= 32'h028B0D43;
    angle_rom[5] <= 32'h0145D7E1;
    angle_rom[6] <= 32'h00A2F61F;
    angle_rom[7] <= 32'h00517C55;
    angle_rom[8] <= 32'h0028BE53;
    angle_rom[9] <= 32'h00145F2F;
    angle_rom[10] <= 32'h000A2F98;
    angle_rom[11] <= 32'h000517CC;
    angle_rom[12] <= 32'h00028BE6;
    angle_rom[13] <= 32'h000145F3;
    angle_rom[14] <= 32'h0000A2FA;
    angle_rom[15] <= 32'h0000517D;
    angle_rom[16] <= 32'h000028BE;
    angle_rom[17] <= 32'h0000145F;
    angle_rom[18] <= 32'h00000A30;
    angle_rom[19] <= 32'h00000518;
    angle_rom[20] <= 32'h0000028C;
  end
else
begin
    angle_rom[0] <= angle_rom[0];
    angle_rom[1] <= angle_rom[1];
    angle_rom[2] <= angle_rom[2];
    angle_rom[3] <= angle_rom[3];
    angle_rom[4] <= angle_rom[4];
    angle_rom[5] <= angle_rom[5];
    angle_rom[6] <= angle_rom[6];
    angle_rom[7] <= angle_rom[7];
    angle_rom[8] <= angle_rom[8];
    angle_rom[9] <= angle_rom[9];
    angle_rom[10] <= angle_rom[10];
    angle_rom[11] <= angle_rom[11];
    angle_rom[12] <= angle_rom[12];
    angle_rom[13] <= angle_rom[13];
    angle_rom[14] <= angle_rom[14];
    angle_rom[15] <= angle_rom[15];
    angle_rom[16] <= angle_rom[16];
    angle_rom[17] <= angle_rom[17];
    angle_rom[18] <= angle_rom[18];
    angle_rom[19] <= angle_rom[19];
    angle_rom[20] <= angle_rom[20];
end

//////////////////////////////////////////////////////////////////
/////////////////////SUPPORT SIGNALS//////////////////////////////
//////////////////////////////////////////////////////////////////
`ifdef CORDIC_16
shortint X_next;
shortint Y_next;
shortint X_after_first_adder;
shortint Y_after_first_adder;
shortint X_after_second_adder;
shortint Y_after_second_adder;
shortint X_mux_one;
shortint Y_mux_one;
shortint X_after_reg;
shortint Y_after_reg;
shortint X_after_shift;
shortint Y_after_shift;
shortint X_after_shift_vect;
shortint Y_after_shift_vect;
`elsif CORDIC_8
byte X_next;
byte Y_next;
byte X_after_first_adder;
byte Y_after_first_adder;
byte X_after_second_adder;
byte Y_after_second_adder;
byte X_mux_one;
byte Y_mux_one;
byte X_after_reg;
byte Y_after_reg;
byte X_after_shift;
byte Y_after_shift;
byte X_after_shift_vect;
byte Y_after_shift_vect;
`else
int X_next;
int Y_next;
int X_after_first_adder;
int Y_after_first_adder;
int X_after_second_adder;
int Y_after_second_adder;
int X_mux_one;
int Y_mux_one;
int X_after_reg;
int Y_after_reg;
int X_after_shift;
int Y_after_shift;
int X_after_shift_vect;
int Y_after_shift_vect;
`endif

logic [WIDTH-1:0] Z_next;
logic [WIDTH-1:0] Z_mux_one;
logic [WIDTH-1:0] Z_after_reg;

//////////////////////////////////////////////////////////////////
/////////////////////SIGN SUPPORT SIGNALS/////////////////////////
//////////////////////////////////////////////////////////////////
logic Z_sign;
logic Z_sign_not;
logic Y_sign;
logic Y_sign_2;

logic sign_for_x_adder;
logic sign_for_y_adder;

logic sign_for_x_adder_2;
logic sign_for_y_adder_2;

//////////////////////////////////////////////////////////////////
///////////////////////Z and Y sign///////////////////////////////
//////////////////////////////////////////////////////////////////
//Z sign
assign Z_sign     = Z_after_reg[WIDTH-1];
assign Z_sign_not = ~Z_after_reg[WIDTH-1];

`ifdef CORDIC_16
assign Y_sign     = Y_after_reg[15];
assign Y_sign_2   = Y_after_first_adder[15];
`elsif CORDIC_8
assign Y_sign     = Y_after_reg[7];
assign Y_sign_2   = Y_after_first_adder[7];
`else
assign Y_sign     = Y_after_reg[31];
assign Y_sign_2   = Y_after_first_adder[31];
`endif

assign sign_for_x_adder = op ? Y_sign : Z_sign;
assign sign_for_y_adder = op ? Y_sign : Z_sign_not;

assign sign_for_x_adder_2 = op ? Y_sign_2 : Z_sign;
assign sign_for_y_adder_2 = op ? Y_sign_2 : Z_sign_not;

//////////////////////////////////////////////////////////////////
///////////////////////X and Y shift//////////////////////////////
//////////////////////////////////////////////////////////////////
assign X_after_shift = (X_after_reg >>> counter);
assign Y_after_shift = (Y_after_reg >>> counter);

assign X_after_shift_vect = (X_after_first_adder >>> counter);
assign Y_after_shift_vect = (Y_after_first_adder >>> counter);

//////////////////////////////////////////////////////////////////
///////////////////////Z stage////////////////////////////////////
//////////////////////////////////////////////////////////////////

//Top mux
assign Z_mux_one = step ? Z_next : Z_i;

//Register
always_ff @(posedge clk or negedge rst)
  //if(~rst)
  if(!rst)
    begin
     Z_after_reg <= 0;
    end
  else if(en_reg == 1'b1)
    begin
     Z_after_reg <= Z_mux_one; 
    end

//Z adder
cordic_addsub_z #(WIDTH)
cordic_addsub_z (.A_i(Z_after_reg),
                 .B_i(angle_rom[counter][31:32-WIDTH]),
                 .sel(Z_sign),
                 .C_o(Z_next)
);

//////////////////////////////////////////////////////////////////
///////////////////////X stage////////////////////////////////////
//////////////////////////////////////////////////////////////////

//Top mux
assign X_mux_one = step ? X_next : X_i;

//Register
always_ff @(posedge clk or negedge rst)
  //if(~rst)
  if(!rst)
    begin
     X_after_reg <= 0;
    end
  else if(en_reg == 1'b1)
    begin
     X_after_reg <= X_mux_one; 
    end

//X first adder
cordic_addsub         #()
cordic_addsub_x_first  (.A_i(X_after_reg),
                        .B_i(Y_after_shift),
                        .sel(sign_for_x_adder),
                        .C_o(X_after_first_adder)
);

//X second adder
cordic_addsub          #()
cordic_addsub_x_second  (.A_i(X_after_first_adder),
                         .B_i(Y_after_shift_vect),
                         .sel(sign_for_x_adder_2),
                         .C_o(X_after_second_adder)
);

assign X_next = conv ? X_after_second_adder : X_after_first_adder;

//////////////////////////////////////////////////////////////////
///////////////////////Y stage////////////////////////////////////
//////////////////////////////////////////////////////////////////

//Top mux
assign Y_mux_one = step ? Y_next : Y_i;

//Register
always_ff @(posedge clk or negedge rst)
  //if(~rst)
  if(!rst)
    begin
     Y_after_reg <= 0;
    end
  else if(en_reg == 1'b1)
    begin
     Y_after_reg <= Y_mux_one; 
    end

//Y first adder
cordic_addsub         #()
cordic_addsub_y_first  (.A_i(Y_after_reg),
                        .B_i(X_after_shift),
                        .sel(sign_for_y_adder),
                        .C_o(Y_after_first_adder)
);

//Y second adder
cordic_addsub          #()
cordic_addsub_y_second  (.A_i(Y_after_first_adder),
                         .B_i(X_after_shift_vect),
                         .sel(sign_for_y_adder_2),
                         .C_o(Y_after_second_adder)
);

assign Y_next = conv ? Y_after_second_adder : Y_after_first_adder;

//////////////////////////////////////////////////////////////////
///////////////////////OUTPUT ASSIGNMENT//////////////////////////
//////////////////////////////////////////////////////////////////
assign X_o = X_next;
assign Y_o = Y_next;
assign Z_o = Z_next;

endmodule




