module cordic_stage_rep_unfolded
#(
  parameter integer unsigned ID = 1,
  parameter integer unsigned WIDTH = 32
)
(
  // Group: clk and enable inputs
  // Inputs:
  //    clk - clock signal
  //    en  - enable signal
  //    rst - reset signal
  input logic                clk,  
  input logic                en_i,
  input logic                rst,
  input logic                neg_i,
  input logic                operation_i,
  // Group: X, Y and Z inputs
  // Inputs:
  //    X - x position used for cosine
  //    Y - y position used for sine
  //    Z - used to determine the next rotation
  //    angle - used to determine the next rotation
  `ifdef CORDIC_16
  input shortint             X_i,  
  input shortint             Y_i,
  `elsif CORDIC_8
  input byte                 X_i,  
  input byte                 Y_i,
  `else
  input int                  X_i,  
  input int                  Y_i,
  `endif
  input logic [WIDTH-1:0]    Z_i,
  input logic [WIDTH-1:0]    angle,
  // Group: X, Y and Z outputs
  // Outputs:
  //    X - x position to the next stage
  //    Y - y position to the next stage
  //    Z - z to the next stage
  `ifdef CORDIC_16
  output shortint            X_o,  
  output shortint            Y_o, 
  `elsif CORDIC_8
  output byte                X_o,  
  output byte                Y_o, 
  `else
  output int                 X_o,  
  output int                 Y_o,
  `endif
  output logic [WIDTH-1:0]   Z_o,
  output logic               en_o,
  output logic               neg_o,
  output logic               operation_o   
);


//rotation direction signals
logic Y_sign, Y_sign_vect;
//post adders signals
`ifdef CORDIC_16
shortint X_next, Y_next; 
shortint X_rot, Y_rot;
shortint X_vect2, Y_vect2;  
shortint X_vect, Y_vect; 
assign Y_sign = (Y_i[15]);
`elsif CORDIC_8
byte X_next, Y_next; 
byte X_rot, Y_rot; 
byte X_vect2, Y_vect2; 
byte X_vect, Y_vect; 
assign Y_sign = (Y_i[7]);
`else
int X_next, Y_next; 
int X_rot, Y_rot; 
int X_vect2, Y_vect2; 
int X_vect, Y_vect; 
assign Y_sign = (Y_i[31]);
`endif

logic [WIDTH-1:0] Z_next;

//rotation direction signals
logic Z_sign;
logic Z_sign_not;
assign Z_sign = Z_i[WIDTH-1];
assign Z_sign_not = ~(Z_i[WIDTH-1]);

//shifted inputs
`ifdef CORDIC_16
shortint X_shift, Y_shift;
shortint X_shift_vect, Y_shift_vect;
assign Y_sign_vect = (Y_shift_vect[15]);
`elsif CORDIC_8
byte X_shift, Y_shift;
byte X_shift_vect, Y_shift_vect;
assign Y_sign_vect = (Y_shift_vect[7]);
`else
int X_shift, Y_shift;
int X_shift_vect, Y_shift_vect;
assign Y_sign_vect = (Y_shift_vect[31]);
`endif

assign X_shift = X_i >>> ID;
assign Y_shift = Y_i >>> ID;

assign X_shift_vect = X_vect2 >>> ID;
assign Y_shift_vect = Y_vect2 >>> ID;

cordic_addsub       #()
cordic_addsub_x_rot (.A_i(X_i),
                     .B_i(Y_shift),
                     .sel(Z_sign),
                     .C_o(X_rot)
);

cordic_addsub       #()
cordic_addsub_y_rot (.A_i(Y_i),
                     .B_i(X_shift),
                     .sel(Z_sign_not),
                     .C_o(Y_rot)
);

cordic_addsub         #()
cordic_addsub_x_vect2 (.A_i(X_i),
                       .B_i(Y_shift),
                       .sel(Y_sign),
                       .C_o(X_vect2)
);

cordic_addsub         #()
cordic_addsub_y_vect2 (.A_i(Y_i),
                       .B_i(X_shift),
                       .sel(Y_sign),
                       .C_o(Y_vect2)
);

cordic_addsub        #()
cordic_addsub_x_vect (.A_i(X_vect2),
                      .B_i(Y_shift_vect),
                      .sel(Y_sign_vect),
                      .C_o(X_vect)
);

cordic_addsub        #()
cordic_addsub_y_vect (.A_i(Y_vect2),
                      .B_i(X_shift_vect),
                      .sel(Y_sign_vect),
                      .C_o(Y_vect)
);

cordic_addsub_z #(WIDTH)
cordic_addsub_z (.A_i(Z_i),
                 .B_i(angle),
                 .sel(Z_sign),
                 .C_o(Z_next)
);

assign X_next = operation_i ? X_vect : X_rot;
assign Y_next = operation_i ? Y_vect : Y_rot;

always_ff @(posedge clk or negedge rst)
  if(!rst)
    begin
     X_o <= 0;
     Y_o <= 0;
     Z_o <= {WIDTH{1'b0}};
     neg_o <= 1'b0;     
    end
  else if(en_i == 1'b1)
    begin
     X_o <= X_next;
     Y_o <= Y_next;
     Z_o <= Z_next;
     neg_o <= neg_i;   
    end
 
always_ff @(posedge clk)
begin
  if(!rst)
  begin
    en_o <= 1'b0;
    operation_o <= 1'b0;
  end
  else
  begin
    en_o <= en_i;
    operation_o <= operation_i;
  end 
end

endmodule
    
