module cordic_top_unfolded
#(
  parameter integer unsigned WIDTH = 32,
  parameter integer unsigned DEPTH = 10
)
(
  // Group: clk and enable inputs
  // Inputs:
  //    clk - clock signal
  //    en  - enable signal
  //    rst - reset signal
  input logic                clk,  
  input logic                en,
  input logic                op,
  input logic                rst,
  output logic               en_o,
  // Group: X, Y and Z inputs
  // Inputs:
  //    X - x position used for cosine
  //    Y - y position used for sine
  //    Z - used to determine the next rotation
  `ifdef CORDIC_16
  input shortint             X_i,  
  input shortint             Y_i,
  `elsif CORDIC_8
  input byte                 X_i,  
  input byte                 Y_i,
  `else
  input int                  X_i,  
  input int                  Y_i,
  `endif
  input logic [WIDTH-1:0]    Z_i,
  // Group: X, Y and Z outputs
  // Outputs:
  //    X - x position to post processing
  //    Y - y position to post processing
  //    Z - z to post processing
  `ifdef CORDIC_16
  output shortint            X_o,  
  output shortint            Y_o, 
  `elsif CORDIC_8
  output byte                X_o,  
  output byte                Y_o, 
  `else
  output int                 X_o,  
  output int                 Y_o,
  `endif
  output logic [WIDTH-1:0]   Z_o   
);

//arctan angle look up table
logic [31:0] angle_rom [20:0];

always_ff @(posedge clk or negedge rst)
if(!rst)
  begin
    angle_rom[0] <= 32'h20000000;
    angle_rom[1] <= 32'h12E4051E;
    angle_rom[2] <= 32'h09FB385B;
    angle_rom[3] <= 32'h051111D4;
    angle_rom[4] <= 32'h028B0D43;
    angle_rom[5] <= 32'h0145D7E1;
    angle_rom[6] <= 32'h00A2F61F;
    angle_rom[7] <= 32'h00517C55;
    angle_rom[8] <= 32'h0028BE53;
    angle_rom[9] <= 32'h00145F2F;
    angle_rom[10] <= 32'h000A2F98;
    angle_rom[11] <= 32'h000517CC;
    angle_rom[12] <= 32'h00028BE6;
    angle_rom[13] <= 32'h000145F3;
    angle_rom[14] <= 32'h0000A2FA;
    angle_rom[15] <= 32'h0000517D;
    angle_rom[16] <= 32'h000028BE;
    angle_rom[17] <= 32'h0000145F;
    angle_rom[18] <= 32'h00000A30;
    angle_rom[19] <= 32'h00000518;
    angle_rom[20] <= 32'h0000028C;
  end
else
begin
    angle_rom[0] <= angle_rom[0];
    angle_rom[1] <= angle_rom[1];
    angle_rom[2] <= angle_rom[2];
    angle_rom[3] <= angle_rom[3];
    angle_rom[4] <= angle_rom[4];
    angle_rom[5] <= angle_rom[5];
    angle_rom[6] <= angle_rom[6];
    angle_rom[7] <= angle_rom[7];
    angle_rom[8] <= angle_rom[8];
    angle_rom[9] <= angle_rom[9];
    angle_rom[10] <= angle_rom[10];
    angle_rom[11] <= angle_rom[11];
    angle_rom[12] <= angle_rom[12];
    angle_rom[13] <= angle_rom[13];
    angle_rom[14] <= angle_rom[14];
    angle_rom[15] <= angle_rom[15];
    angle_rom[16] <= angle_rom[16];
    angle_rom[17] <= angle_rom[17];
    angle_rom[18] <= angle_rom[18];
    angle_rom[19] <= angle_rom[19];
    angle_rom[20] <= angle_rom[20];
end

//support signals
`ifdef CORDIC_16
shortint X_support [DEPTH:0];
shortint Y_support [DEPTH:0];
`elsif CORDIC_8
byte X_support [DEPTH:0];
byte Y_support [DEPTH:0];
`else
int X_support [DEPTH:0];
int Y_support [DEPTH:0];
`endif

logic [WIDTH-1:0] Z_support [DEPTH:0];

assign X_support [0] = X_i;
assign Y_support [0] = Y_i;

//final results
`ifdef CORDIC_16
shortint X_final;
shortint Y_final;
`elsif CORDIC_8
byte X_final;
byte Y_final;
`else
int X_final;
int Y_final;
`endif

logic [WIDTH-1:0] Z_final;

//enable support
logic [DEPTH:0] en_support;
assign en_support[0] = en;

//operation support
logic [DEPTH:0] op_support;
assign op_support[0] = op;

//neg support
logic [DEPTH:0] neg_support;

cordic_pre #(WIDTH)
cordic_pre (.Z_i(Z_i),
            .Z_o(Z_support [0][WIDTH-1:0]),
            //.Z_o(Z_support [0][31:32-WIDTH]),
            .neg(neg_support[0]));

//genvar
genvar i;

for (i = 0; i < DEPTH; i++) 
begin
  if(i == 0)
    cordic_stage_first_unfolded #(i, WIDTH)
    cordic_stage_first_unfolded (.clk(clk),
                                 .en_i(en_support[i]),
                                 .rst(rst),
                                 .neg_i(neg_support[i]),
                                 .operation_i(op_support[i]),
                                 .X_i(X_support[i]),
                                 .Y_i(Y_support[i]),
                                 .Z_i(Z_support[i][WIDTH-1:0]),
                                 .angle(angle_rom[i][31:31-WIDTH+1]),
                                 .X_o(X_support[i+1]),
                                 .Y_o(Y_support[i+1]),
                                 .Z_o(Z_support[i+1][WIDTH-1:0]),
                                 .en_o(en_support[i+1]),
                                 .neg_o(neg_support[i+1]),
                                 .operation_o(op_support[i+1]));
  else if(i == 4 || i == 13)
    cordic_stage_rep_unfolded #(i, WIDTH)
    cordic_stage_rep_unfolded (.clk(clk),
                               .en_i(en_support[i]),
                               .rst(rst),
                               .neg_i(neg_support[i]),
                               .operation_i(op_support[i]),
                               .X_i(X_support[i]),
                               .Y_i(Y_support[i]),
                               .Z_i(Z_support[i][WIDTH-1:0]),
                               .angle(angle_rom[i][31:31-WIDTH+1]),
                               .X_o(X_support[i+1]),
                               .Y_o(Y_support[i+1]),
                               .Z_o(Z_support[i+1][WIDTH-1:0]),
                               .en_o(en_support[i+1]),
                               .neg_o(neg_support[i+1]),
                               .operation_o(op_support[i+1]));
  else
    cordic_stage_unfolded #(i, WIDTH)
    cordic_stage_unfolded (.clk(clk),
                           .en_i(en_support[i]),
                           .rst(rst),
                           .neg_i(neg_support[i]),
                           .operation_i(op_support[i]),
                           .X_i(X_support[i]),
                           .Y_i(Y_support[i]),
                           .Z_i(Z_support[i][WIDTH-1:0]),
                           .angle(angle_rom[i][31:31-WIDTH+1]),
                           .X_o(X_support[i+1]),
                           .Y_o(Y_support[i+1]),
                           .Z_o(Z_support[i+1][WIDTH-1:0]),
                           .en_o(en_support[i+1]),
                           .neg_o(neg_support[i+1]),
                           .operation_o(op_support[i+1]));
end

cordic_post #(WIDTH)
cordic_post (.en_i(en_support[DEPTH]),
             .X_i(X_support [DEPTH]),
             .Y_i(Y_support [DEPTH]),
             .Z_i(Z_support [DEPTH][WIDTH-1:0]),
             .neg_i(neg_support[DEPTH]),
             .operation_i(op_support[DEPTH]),
             .X_o(X_final),
             .Y_o(Y_final),
             .Z_o(Z_final),
             .en_o(en_o)
);

assign X_o = X_final;
assign Y_o = Y_final;
assign Z_o = Z_final;

endmodule

